import pytest
from pfylter.core import AllFilters, AnyFilter, AbstractFilter, NotFilter, InFilter
from pfylter.strings import LenFilter, StartsWithFilter, ContainsFilter, RegexFilter, LengthRangeFilter

class DummyFilter(AbstractFilter[int]):
    def keep(self, instance: int) -> bool:
        return instance > 5

class EvenFilter(AbstractFilter[int]):
    def keep(self, instance: int) -> bool:
        return instance % 2 == 0

@pytest.fixture
def values():
    return [1, 2, 3, 6, 7, 8]

def test_all_filters(values):
    all_filters = AllFilters([DummyFilter(), EvenFilter()])
    assert all_filters.apply(values) == [6, 8]

def test_any_filter(values):
    any_filter = AnyFilter([DummyFilter(), EvenFilter()])
    assert any_filter.apply(values) == [2, 6, 7, 8]

def test_in_filter(values):
    in_filter = InFilter([2, 4, 20])
    assert in_filter.apply(values) == [2]

@pytest.fixture
def string_values():
    return ['A', 'ABCD', 'B', 'BCDE', 'C', 'AAAAAAA']

def test_contains_filter(string_values):
    f = ContainsFilter('BC')
    assert f.apply(string_values) == ['ABCD', 'BCDE']

def test_len_filter(string_values):
    f = LenFilter(1)
    assert f.apply(string_values) == ['A', 'B', 'C']

def test_length_range_filter(string_values):
    f = LengthRangeFilter(3,5)
    assert f.apply(string_values) == ['ABCD', 'BCDE']

def test_not_len_filter(string_values):
    f = NotFilter(LenFilter(1))
    assert f.apply(string_values) == ['ABCD', 'BCDE', 'AAAAAAA']

def test_all_filters_string(string_values):
    f = AllFilters([LenFilter(4), StartsWithFilter('A')])
    assert f.apply(string_values) == ['ABCD']

def test_any_filters_string(string_values):
    f = AnyFilter([LenFilter(4), StartsWithFilter('A')])
    assert f.apply(string_values) == ['A', 'ABCD', 'BCDE', 'AAAAAAA']

def test_complex_any_all_filters_string(string_values):
    f = AnyFilter([
        AllFilters([LenFilter(4), StartsWithFilter('A')]),
        AllFilters([LenFilter(1), StartsWithFilter('B')])
    ])
    assert f.apply(string_values) == ['ABCD', 'B']

def test_all_filters_with_any(string_values):
    f = AllFilters([
        StartsWithFilter('A'),
        AnyFilter([LenFilter(1), LenFilter(4)])
    ])
    assert f.apply(string_values) == ['A', 'ABCD']

def test_exclude_bc_or_len1_allfilters(string_values):
    f = AllFilters([NotFilter(LenFilter(1)), NotFilter(ContainsFilter('BC'))])
    assert f.apply(string_values) == ['AAAAAAA']

def test_exclude_bc_or_len1_anyfilter(string_values):
    f = NotFilter(AnyFilter([ContainsFilter('BC'), LenFilter(1)]))
    assert f.apply(string_values) == ['AAAAAAA']

def test_regex_filter(string_values):
    f = RegexFilter('^A.*[D|A]$')
    assert f.apply(string_values) == ['ABCD', 'AAAAAAA']


from abc import ABC, abstractmethod
from typing import Generic, TypeVar, List, Callable, Set

T = TypeVar('T')

class AbstractFilter(ABC, Generic[T]):
    def apply(self, elements: List[T]) -> List[T]:
        return [e for e in elements if self.keep(e)]

    @abstractmethod
    def keep(self, instance: T) -> bool:
        pass


class AllFilters(AbstractFilter[T]):
    def __init__(self, filters: List[AbstractFilter[T]]) -> None:
        self.filters = filters

    def keep(self, instance: T) -> bool:
        return all(f.keep(instance) for f in self.filters)


class AnyFilter(AbstractFilter[T]):
    def __init__(self, filters: List[AbstractFilter[T]]) -> None:
        self.filters = filters

    def keep(self, instance: T) -> bool:
        return any(f.keep(instance) for f in self.filters)


class NotFilter(AbstractFilter[T]):
    def __init__(self, base_filter: AbstractFilter[T]) -> None:
        self.base_filter = base_filter

    def keep(self, instance: T) -> bool:
        return not self.base_filter.keep(instance)


class LambdaFilter(AbstractFilter[T]):
    def __init__(self, func: Callable[[T], bool]) -> None:
        self.func = func

    def keep(self, instance: T) -> bool:
        return self.func(instance)


class InFilter(AbstractFilter[T]):
    def __init__(self, options: Set[T] | List[T]) -> None:
        self.options = set(options)

    def keep(self, instance: T) -> bool:
        return instance in self.options


if __name__ == '__main__':
    example = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    print('Numbers greater than 5:')
    print(LambdaFilter(lambda x: x > 5).apply(example))

    print('Numbers equal or lower than 5:')
    print(NotFilter(LambdaFilter(lambda x: x > 5)).apply(example))
    
    print('Numbers greater than 5 and divisible by two:')
    print(AllFilters([LambdaFilter(lambda x: x > 5), LambdaFilter(lambda x: x % 2 == 0)]).apply(example))
    
    print('Numbers greater than 5 and divisible by three:')
    print(AllFilters([LambdaFilter(lambda x: x > 5), LambdaFilter(lambda x: x % 3 == 0)]).apply(example))
    
    print('Numbers greater than 5 or divisible by two:')
    print(AnyFilter([LambdaFilter(lambda x: x > 5), LambdaFilter(lambda x: x % 2 == 0)]).apply(example))
    
    print('Numbers in [2, 4, 20]:')
    print(InFilter([2, 4, 20]).apply(example))



import re
from pfylter.core import AbstractFilter, AllFilters, AnyFilter, NotFilter

class LenFilter(AbstractFilter[str]):
        def __init__(self, length: int) -> None:
            self.length = length

        def keep(self, instance: str) -> bool:
            return len(instance) == self.length


class LengthRangeFilter(AbstractFilter[str]):
    def __init__(self, min_len: int, max_len: int) -> None:
        self.min_len = min_len
        self.max_len = max_len

    def keep(self, instance: str) -> bool:
        return self.min_len <= len(instance) <= self.max_len


class StartsWithFilter(AbstractFilter[str]):
    def __init__(self, start: str) -> None:
        self.start = start

    def keep(self, instance: str) -> bool:
        return instance.startswith(self.start)


class EndsWithFilter(AbstractFilter[str]):
    def __init__(self, suffix: str) -> None:
        self.suffix = suffix

    def keep(self, instance: str) -> bool:
        return instance.endswith(self.suffix)


class ContainsFilter(AbstractFilter[str]):
    def __init__(self, substring: str) -> None:
        self.substring = substring

    def keep(self, instance: str) -> bool:
        return self.substring in instance


class RegexFilter(AbstractFilter[str]):
    def __init__(self, pattern: str) -> None:
        self.pattern = re.compile(pattern)

    def keep(self, instance: str) -> bool:
        return bool(self.pattern.search(instance))

if __name__ == '__main__':
    example = ['A', 'ABCD', 'B', 'BCDE', 'C', 'AAAAAAA']

    print('Strings containing "BC":')
    print(ContainsFilter('BC').apply(example))

    print('Strings with length one:')
    print(LenFilter(1).apply(example))

    print('Strings with length different than one:')
    print(NotFilter(LenFilter(1)).apply(example))
    
    print('Strings with length four and starting with "A":')
    print(AllFilters([LenFilter(4), StartsWithFilter('A')]).apply(example))

    print('Strings with length four or starting with "A":')
    print(AnyFilter([LenFilter(4), StartsWithFilter('A')]).apply(example))

    print('Strings with length four and starting with "A" or length one and starting with "B":')
    print(AnyFilter([
            AllFilters([LenFilter(4), StartsWithFilter('A')]),
            AllFilters([LenFilter(1), StartsWithFilter('B')])
        ]).apply(example))

    print('Strings starting with "A" and either length one or length four:')
    print(AllFilters([
           StartsWithFilter("A"), 
           AnyFilter([LenFilter(1), LenFilter(4)])
        ]).apply(example))

    print('Exclude any string that includes BC or has length 1 (using AllFilters):')
    print(AllFilters([NotFilter(LenFilter(1)), NotFilter(ContainsFilter('BC'))]).apply(example))

    print('Exclude any string that includes BC or has length 1 (using AnyFilter):')
    print(NotFilter(AnyFilter([ContainsFilter('BC'), LenFilter(1)])).apply(example))

    print('Strings that start with A followed by any chars and ends in D or A:')
    print(RegexFilter('^A.*[D|A]$').apply(example))
